!******************************************
! Creates a background image for omdetect
!*****************************************
module backgroundimage_mod

use dal
use cal
use omskybackground_mod
use omutility_mod
use omsmoothimage_mod
use omsourceparam_mod
use ominterpolate_mod
use cubicshepard_mod
use gridinterpolation_mod
use omsourcemap_mod
use parameters_mod
implicit none



interface createBackgroundImage
    module procedure  createBackgroundImage1,  &
                      createBackgroundImage2,  &
                      createBackgroundImage3,  &
                      createBackgroundImage4,  &
                      createBackgroundImage5,  &
                      createBackgroundImage6

end interface

integer(int8), private :: skip = 2
contains
!*****************************************************
!
!*****************************************************
subroutine setBackgroundSkip( iSkip )
    integer(int8), intent(in) :: iSkip

    skip = iskip


end subroutine setBackgroundSkip

!******************************************************
!
!******************************************************
subroutine createBackgroundImage1(image, qualityImage, &
                                  backgroundimage, backgroundsigmaimage)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
  


    call createBackgroundImage2(image, qualityImage, 10000_int16, &
                                backgroundimage=backgroundimage, backgroundsigmaimage=backgroundsigmaImage)

end subroutine createBackgroundImage1


!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage2(image, qualityImage, width, &
               backgroundImage, backgroundsigmaImage, maxheight)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
 
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    integer(int16), pointer :: levelImage(:, :) => null()
    logical, pointer :: pixelOkay(:, :) => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    logical, pointer :: computed(:, :) => null()

   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = 50
    dw1 = dw - 1
    
! write(*, *) "width=", width, " dx=", dx, " d6y=", dy
    levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")

    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if

    deallocate(LevelImage)
    nullify(levelImage)
     
    if (.not. associated(backgroundimage)) then
    

        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if




    backgroundImage = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.







    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.





    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, computed, &
               dx, dy, dr, mode)

        end do


    end do


    deallocate(pixelOkay)
    nullify(pixelOkay)





    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( backgroundimage, backgroundsigmaImage, computed)


    deallocate(computed)
    nullify(computed)



    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))


!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage2

!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage5(image, width, maxheight)
    real(single), pointer :: image(:, :)
   
    integer(int16), intent(in) :: width
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
 
    logical, pointer, dimension(:, :) :: pixelOkay => null(), computed => null()
    real(single), pointer :: tempImage1(:, :) => null()
    real(single), pointer :: tempImage2(:, :) => null()

  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    


    tempImage1 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage1")
    tempImage2 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage2")

    tempimage1 = image



   
 




    mode = ComputeBackgroundMode2(image, nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 3.0 * max(1.0, sigma)
    end if
   
     

    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
   



    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.




  

    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               tempimage1, tempimage2, pixelOkay, computed, &
               dx, dy, dr, mode)

        end do


    end do


    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( tempimage1, tempImage2, computed)



    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  


    image = tempimage1
    deallocate(pixelOkay, tempimage1, tempimage2, computed)
    nullify(pixelOkay, tempimage1, tempimage2, computed)

    call smoothimage(image, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage5

!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage6(image, levelImage, width, maxheight)
    real(single), pointer :: image(:, :)
    integer(int16), pointer :: levelImage(:, :)

    integer(int16), intent(in) :: width
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
  
    logical, pointer, dimension(:, :) :: pixelOkay => null(), computed => null()
    real(single), pointer :: tempImage1(:, :) => null()
    real(single), pointer :: tempImage2(:, :) => null()

  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)




    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    


    tempImage1 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage1")
    tempImage2 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage2")

    tempimage1 = image

    mode = ComputeBackgroundMode(image, levelimage, nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 3.0 * max(1.0, sigma)
    end if
    
     

    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. levelimage .eq. 0_int16)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
   
    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.

   
    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               tempimage1, tempimage2, pixelOkay, computed, &
               dx, dy, dr, mode)

        end do


    end do




    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( tempimage1, tempImage2, computed)



    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  


    image = tempimage1
    deallocate(computed, pixelOkay, tempimage1, tempimage2)
    nullify(computed, pixelOkay, tempimage1, tempimage2)

    call smoothimage(image, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))


!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage6



!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage3(image, width, &
               backgroundImage, backgroundsigmaImage, maxheight)
    real(single), pointer :: image(:, :)
  
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    integer(int16), pointer :: levelImage(:, :) => null()
    logical, pointer, dimension(:, :) :: pixelOkay => null(), computed => null()
    logical(boolean), pointer :: qualityImage(:, :)
 
 
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
  
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    


  


    allocate(qualityImage(0 : nx, 0 : ny))

    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw1 = dw - 1
 

    where(image .lt. 0.0) 
        qualityImage = .false.
    elsewhere
        qualityImage = .true.
    end where

    levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")
    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    deallocate(LevelImage)
    nullify(levelImage)



     
    if (associated(backgroundimage)) deallocate(backgroundimage)
    if (associated(backgroundSigmaimage)) deallocate(backgroundSigmaimage)

    backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")



    backgroundImage = image


    allocate(pixelOkay(0 : nx, 0 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.





 
    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.






    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, computed, &
               dx, dy, dr, mode)

        end do


    end do




    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( backgroundimage, backgroundsigmaImage, computed)









    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

  
    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))

  


    deallocate(computed, pixelOkay, qualityImage)
    nullify(computed, pixelOkay, qualityImage)



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fits", "background image")
!call fatal("stopped displaying modeimage.fits")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage3


!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage4(image, qualityImage, width, &
               backgroundImage, backgroundsigmaImage,         &
               levelimage, maxheight)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
 
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    integer(int16), pointer :: levelImage(:, :) 
 
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    logical, pointer, dimension(:, :) :: computed => null(), pixelOkay => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    


 


    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    

    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    
     
    if (.not. associated(backgroundimage)) then
    

        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if


    backgroundImage = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage .and. levelimage .eq. 0)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.



  
    allocate(computed( 0 : nx, 0 : ny) )
    computed = .false.




    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, computed,&
               dx, dy, dr, mode)

        end do


    end do




    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( backgroundimage, backgroundsigmaImage, computed)






    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

  
 
    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))






    deallocate(computed, pixelOkay)
    nullify(computed, pixelOkay)



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage4


!***************************************************************
!
!**************************************************************
subroutine CreateBackgroundStrip(il, iu, jl, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               computed, dx, dy, dr, mode)

    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)

  
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    logical, pointer, dimension(:, :) :: pixelOkay, computed
    integer(int16), intent(in) :: dx, dy, dr
    real(single), intent(inout) :: mode

    integer(int16) :: i, j, l, k
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int32) :: count, pixelSize, n   

    logical, pointer :: usepixel(:) => null()
    real(single) :: sigma, mode1 
    real(single), pointer :: pixelValue(:) => null()
    integer, pointer :: pixelIndex(:, :) => null()
  
  

!write(*, *) "dx=", dx, " dy=", dy, " mode=", mode, " dr=", dr
    !***************************************************************
    !
    !
    !**************************************************************
    nx = ubound(image, 1)
    ny = ubound(image, 2)

 

    il1 = il - dr
    if (il1 < 0) then
        il1 = 0
    end if
    iu1 = iu + dr
    if (iu1 > nx) then
        iu1 = nx
         
    end if 
  

    jl1 = jl - dr
    if (jl1 < 0) then
        jl1 = 0
    end if
    ju1 = ju + dr
    if (ju1 > ny) then
        ju1 = ny
    

    end if

!write(*, *) "Line 895 il1=", il1, " iu1=", iu1, " jl1=", jl1, " ju1=", ju1



    call SetUpIndices(il1, iu1, jl1, ju1, image, pixelOkay, pixelValue, pixelIndex)

!write(*, *) "Line 901"

    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    usePixel = .false.




    mode1 = mode !+ 5.0 * sqrt(mode)
   


!write(*, *) "Line 914 createbackgroudnstrip mode1=", mode1, &
!            " il=", il, " iu=", iu, " jl=", jl, " ju=", ju



    ! Create the background image
    do j = jl, ju, skip        






        jl1 = j - dy
        if (jl1 < 0) then
            jl1 = 0
            ju1 = min(ny, dr)
        else 
            ju1 = j + dy
            if (ju1 > ny) then
                ju1 = ny
                jl1 = max(0_int16, ny - dr)
            end if 
  
        end if
      





!write(*, *) "j=", j, " jl1=", jl1, " ju1=", ju1, " il=", il, " iu=", iu  
       
        do i = il, iu, skip

!            if (mod((i + 1) *(j + 1) , 2) .ne. 0) cycle
            if (image(i, j) .lt. 0.0) then ! Ignore negative image values

                backgroundimage(i, j) = -1.0
                backgroundsigmaimage(i, j) = 1000.0
                computed(i, j ) = .true.
       
                cycle
            end if
            computed(i, j ) = .true.
           

          
            il1 = i - dx
            if (il1 < 0) then
                il1 = 0
                iu1 = min(nx, dr)
            else 
                iu1 = i + dx
                if (iu1 > nx) then
                    iu1 = nx
                    il1 = max(0_int16, nx - dr)
                end if
            end if 
  

                
            count = 0           
     
            do l = jl1, ju1

                do k = il1, iu1

                    if (pixelOkay(k, l)) then
                        count = count + 1
                        
                        n = pixelIndex(k, l)
                        usePixel(n) = .true.
                      
                    end if
                end do
            end do     


            if (count .gt. 10) then  ! Was 10

                backgroundimage(i, j) = BackgroundMode11(count, usePixel, pixelValue, &
                        0.2, 2.0, sigma) 

 
                mode1 = backgroundimage(i, j)        
               

                backgroundsigmaimage(i, j) = max(1.0, sigma)
!if (abs(i - 464) < 20 .and. abs(j - 553) < 20) then
!       write(*, *) "i=", i, " j=", j, " mode1=", mode1, " count=", count
!end if
           else

                backgroundimage(i, j) = mode1 ! image(i, j)
                backgroundsigmaimage(i, j) = max(1.0, sqrt(mode1) )
                usePixel = .false.



!write(*, *) "i=", i, " j=", j, " sqrtmode=", sqrtMode
!call fatal("stopped at line 1013")

 
            end if
!write(*, *) "Line 973 count=", count, " backgroundimage=", backgroundimage(i, j)
!write(*, *) "il1=", il1, " iu1=", iu1, " jl1=", jl1, " ju1=", ju1

    
        end do




    end do







! call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "backgroundimage.fit", "background image")

!call saveImage( backgroundsigmaimage, nx + 1_int16, ny + 1_int16, "sigmaimage.fit", "background image")


!  call fatal("stopped at line 1090 displaying backgroundimage.fit and sigmaimage.fit")  



    mode = mode1

    deallocate(usePixel, pixelValue, pixelIndex)
    nullify(usePixel, pixelValue, pixelIndex)

  
 
end subroutine CreateBackgroundStrip



!***************************************************************
!
!**************************************************************
subroutine createGrismBackgroundImage(image, dx1, dy1, &
               backgroundImage, backgroundsigmaImage,                &
               levelimage, maxheight)
    real(single), pointer :: image(:, :)
  
  
    integer(int16), intent(in) :: dx1, dy1
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    integer(int16), pointer :: levelImage(:, :) 
 
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
    
    real(single) :: globalMedian, sigma, mode
    logical, pointer, dimension(:, :) :: computed => null(), pixelOkay => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1, width
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
 
    dx = dx1
    dy = dy1

    width = max(dx, dy) 
    dr = min(width, nx)
    dr = min(dr, ny)

    

    dy = dx1 / 2
    dx = min(nx, dx)
    dy = dy1 / 2
    dy = min(ny, dy)
    

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    

    mode = ComputeBackgroundMode(image, levelImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    
     
    if (.not. associated(backgroundimage)) then
    

        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if


    backgroundImage = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. levelimage .eq. 0 .and. image .ge. 0.0)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.




    allocate(computed(0 : nx, 0 : ny))
    computed = .false.


    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            !**************************************************************
            ! The box size used in computing the background is 2 x dx + 1 by 2 x dy + 1             
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, computed, &
               dx, dy, dr, mode)

        end do


    end do







    !****************************************************************************
    ! Fill in the missing background pixels by interpolation
    !***************************************************************************
    call FillInBackgroundPixels( backgroundimage, backgroundsigmaImage, computed)


    deallocate(computed)
    nullify(computed)





    


    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  








    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(6 / binx, int32), int(6 / binx, int32))


    

    where(image .le. 0.0) backgroundimage = 0.0
    deallocate(pixelOkay)
    nullify(pixelOkay)



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createGrismBackgroundImage



!************************************************
!
!***********************************************
real(single) Function PixelAverage(i, j, image1, image2, pixel, dx, computed, fb)
    integer(int16), intent(in) :: i, j
    real(single), pointer, dimension(:, :) :: image1, image2
    type(pixelStruct), pointer, dimension(:) :: pixel

    logical, pointer, dimension(:, :) :: computed
  

    integer(int8), intent(in) :: dx
    real(single), intent(out) :: fb


    ! Local variables     
    integer(int16) :: nx, ny, il, iu, jl, ju, i1, j1
    integer(int32) :: k   
    real :: xPos, yPos


    

    nx = ubound(image1, 1)
    ny = ubound(image1, 2)
    
    il = max(0_int16, i - dx)
    iu = min(nx, i + dx)
    jl = max(0_int16, j - dx)
    ju = min(ny, j + dx)



    k = 0
   

    do i1 = il, iu
        do j1 = jl, ju
            
            if (i1 .eq. i .and. j1 .eq. j) cycle
            if (.not. computed(i1, j1)) cycle 
            if (image1(i1, j1) .lt. 0.0) cycle
            k = k + 1
            pixel(k) = pixelStruct( real(i1), real(j1), 0, image1(i1, j1), image2(i1, j1), 0)  
            

         end do
    end do  
  

    !***************************************************************************
    ! Interpolate to get the background counts
    !**************************************************************************

    xPos = real(i)
    yPos = real(j)
  
    PixelAverage = neighbourInterpolate( xPos, yPos, k, real(pixel%x), real(pixel%y), pixel%counts ) 
          

  
    !***************************************************************************
    ! Interpolate to get the background counts
    !**************************************************************************
    fb = neighbourInterpolate( xPos, yPos, k, real(pixel%x), real(pixel%y), pixel%backgroundcounts ) 
          



    

end Function PixelAverage

!*****************************************************************
!
!****************************************************************
subroutine initialiseMedian(image, pixelOkay, pixelIndex, pixelValue, usePixel)
    real(single), pointer :: image(:, :)
    logical, pointer :: pixelOkay(:, :)
    integer, pointer :: pixelIndex(:)
    real(single), pointer :: pixelValue(:)
    logical, pointer :: usePixel(:, :)
  
    integer :: i, j, k, nx, ny

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    k = count(pixelOkay)
    pixelValue => allocateSingle(1, k, "pixelValue")
    pixelIndex => allocateInt32(1, k, "pixelIndex")
    allocate(usePixel(1, k))

    k = 0
    do i = 0, nx

        do j = 0, ny
            if (pixelOkay(i, j)) then
                k = k + 1
                pixelValue(k) = image(i, j) ! The list of usable pixels for the median
            end if
        end do
    end do
 
    !******************************************************************
    ! Index sort them
    ! The pixelvalue array is not sorted, pixelindex(
    ! will give the position of each value of pixelvalue in
    ! order of increasing order
    !*********************************************************
    call IndexSort(pixelValue, pixelIndex)


end subroutine initialiseMedian

!********************************************************************
! subroutine SetUpIndices
! 
!
!********************************************************************

subroutine SetUpIndices(il, iu, jl, ju, image, usePixel, pixelValue, pixelIndex)
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)
    logical, pointer :: usePixel(:, :)
    real(single), pointer :: pixelValue(:) 
    integer, pointer :: pixelIndex(:, :)
  
    integer(int32), pointer, dimension(:) :: arrayIndex => null()

    integer :: i, j, k
    type(PixelStruct), pointer :: pixel(:) => null()
 

  

    k = count(usePixel(il : iu, jl : ju)) ! The number of pixels to be checked
!write(*, *) "Line 698 k=", k  
    arrayIndex => allocateInt32(1, k, "arrayIndex")

    if (associated(pixelValue)) deallocate(pixelValue)
    if (associated(pixelIndex)) deallocate(pixelIndex)

    pixelValue => allocateSingle(1, k, "pixelValue")
  

    allocate(pixelIndex(il : iu, jl : ju)) 
    allocate(pixel(1 : k))

    k = 0 

    do j = jl, ju
        do i = il, iu
            if (usePixel(i, j)) then
                k = k + 1
                pixel(k)%x = i
                pixel(k)%y = j
                pixel(k)%counts = image(i, j)
 
            end if
        end do
    end do
!write(*, *) "Line 724 k=", k  


    ! Index sort the pixel values- pixelValue will be sorted into 
    ! increasing pixel value

    call IndexSort(pixel%counts, arrayIndex)


    do i = 1, k
        j = arrayIndex(i)
        pixelValue(i) = pixel(j)%counts
        pixelIndex(pixel(j)%x, pixel(j)%y) = i
    end do


    deallocate(pixel, arrayIndex)
    nullify(pixel, arrayIndex)

end subroutine SetUpIndices


!*****************************************
!
! Function BackgroundMode
!
!*****************************************
function BackgroundMode10(nData, data, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(1 : ndata)

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
 
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k
     
    real(single)   :: median
    real(double)   :: variance, testSigma
    real(single)   :: highValue
    integer(int32) :: m, nmin

    nList = nData

    nmin = ishft(ndata, -1)
             
    call HeapSort(data, nData)


  
    ! reject data at more than nSigma from data
   
    testSigma = 0.0
!write(*, *) "background mode 1"
    sigmaClip: do i = 1, 20   !  was 20
      
             
        median = SortedMedian(data, nlist)
               
    


!        write(*, *) "i = ", i, " median=", median, " nlist=", nlist
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
            
        variance = dot_product(data(1 : nlist) - median, data(1 : nlist) - median) 
        sigma = sqrt(variance / dble(nlist - 1))
     
      
        highValue = median + nSigma * sigma

!        lowValue = median - nsigma * sigma

   
        m = 0 
        do k = 1, nlist
!            if(data(k) .le. highValue .and. data(k) .ge. lowValue) then
            if(data(k) .le. highValue) then

  
                m = m + 1
                data(m) = data(k)
               
            end if
        end do
       

        if(m .lt. nmin) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip

  
    
    mode = sum(data(1 : nList)) / dble(nList)
   
    

!write(*, *) "Background mode=", mode, "+/-", sigma, " n=", nlist, " no of iterations=", i  
    
end function BackgroundMode10



!*****************************************
! Function BackgroundMode
!*****************************************
function BackgroundMode11(nData, usePixel, pixelValue, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData

    logical :: usePixel(:)
    real(single) :: pixelValue(: )

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma

    real(single) :: data(1 : nData) 
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k
    real(single)   :: median
    real(double)   :: variance, testSigma, diff
    real(single)   :: highValue, lowValue
    integer(int32) :: m, nmin



    k = 0

    do i = 1, size(usePixel)
        if (usePixel(i)) then
            k = k + 1

            data(k) = pixelValue(i)
        
            usePixel(i) = .false.

        end if
    end do


    nList = nData

    nmin = ishft(ndata, -1)
             
  
    ! reject data at more than nSigma from data
   
    testSigma = 0.0

    sigmaClip: do i = 1, 20   !  was 20
      
             


        !********************************************************************
        ! Get the median
        !*******************************************************************
        k = ishft(nlist, -1)
        if (ishft(k, 1) .eq. nlist) then   ! Even number
            median = (data(k) + data(k + 1)) / 2.0
        else
            median = data(k + 1)
        end if

        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
            
        variance = 0.0_double
        do k = 1, nList
            diff = data(k) - median
            variance = variance + diff * diff 
         
        end do
        sigma = sqrt(variance / dble(nlist - 1))
     
      
        highValue = median + nSigma * sigma

        lowValue = median - nsigma * sigma

   
        m = 0  
        do k = 1, nlist
            if (data(k) .lt. lowValue) cycle ! Put in on Monday
            if(data(k) .gt. highValue) exit
  
            m = m + 1
            data(m) = data(k)
        end do
       

        if(m .lt. nmin .or. sigma .lt. 0.1) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip

  
    
    mode = sum(data(1 : nList)) / dble(nList)
            
    variance = 0.0_double
    do k = 1, nList
        diff = data(k) - mode
        variance = variance + diff * diff 
         
    end do
    sigma = sqrt(variance / dble(nlist - 1))
 
    
end function BackgroundMode11

!*****************************************************************
!
! RefineBackgroundImage
! Interpolates background across extended-source regions
!
!*****************************************************************
subroutine RefineBackgroundImage1(image, levelimage, source, nSources)
    real(single), pointer :: image(:, :)
    integer(int16), pointer :: levelImage(:, :)

    type(sourceParameters), pointer :: source(:)

    integer(int16), intent(in) :: nSources
    integer(int16) :: i, j, k, nx, ny, il, iu, jl, ju, dx
    integer(int16) :: dx1
    integer(int32) :: n
    real(single), pointer, dimension(:) :: x => null(), y => null(), f => null()
    real(single), pointer :: tempImage(:, :)
    real(single) :: xPos, yPos, globalMode, sigma 
    integer(int16), pointer :: tempLevelImage(:, :) => null()
   
    if (nSources .eq. 0) return
    write(errstr,'(a)') "Refining background  image"
    call message(AppMsg, Verbose)


    nx = ubound(image, 1)
    ny = ubound(image, 2)
    tempImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempimage")
    tempLevelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "tempLevelimage")


    tempimage = image

    globalMode = ComputeBackgroundMode3(image, sigma)
write(*, *) "Line 980 global mode=", globalMode
    globalMode = globalMode + 2.0 * sigma


    dx = 40 / binx 
    dx1 = 12 / binx


    n = (maxVal(source%imax - source%imin) + 2 * dx + 1) * &
        (maxVal(source%jmax - source%jmin) + 2 * dx + 1)

    x => allocateSingle(1, n, "x")
    y => allocateSingle(1, n, "y")
    f => allocateSingle(1, n, "f")
  

    do i = 1, nSources

        if (source(i)%extensionFlag .eq. 0) cycle
        if (source(i)%noSourcePixels .lt. 5) cycle

        il = max(0_int16, source(i)%imin - dx)
        iu = min(nx, source(i)%imax + dx)
        jl = max(0_int16, source(i)%jmin - dx)
        ju = min(ny, source(i)%jmax + dx)

        tempLevelImage(il : iu, jl : ju) = levelImage(il : iu, jl : ju)
        call DilateLevelImage2(templevelimage, il, iu, jl, ju, dx1, int(source(i)%number, int16))

        n = 0

        do k = jl, ju
            do j = il, iu
!                if(templevelImage(j, k) .ne. 0) cycle
                if (image(j, k) .gt. globalMode) cycle                
                n = n + 1
                x(n) = real(j) + 0.5
                y(n) = real(k) + 0.5
                f(n) = image(j, k)

            end do
        end do


        if (n .lt. 10) then 
            cycle
        end if

         
        do j = il, iu
            xPos = real(j) + 0.5
       
            do k =  jl, ju
!                if(templevelImage(j, k) .ne. source(i)%number) cycle
 
                if (image(j, k) .le. globalMode) cycle                
      
                yPos = real(k) + 0.5
                tempImage(j, k) = NeighbourInterpolate(xpos, ypos, n, x, y, f)


            end do
        end do
           
          
    end do
100 continue
    deallocate(x, y, f)
    nullify(x, y, f)
        
!call saveImage(image, nx + 1_int16, ny + 1_int16, "image.fits", "background image")
!call saveImage(tempimage, nx + 1_int16, ny + 1_int16, "tempimage.fits", "background image")
!call saveImage(templevelimage, nx + 1_int16, ny + 1_int16, "templevelimage.fits", "background image")

!call fatal("Line 1068 displaying image.fits and tempimage.fits and templevelimage.fits")

    image = tempImage
    deallocate(tempimage, tempLevelImage)
    nullify(tempimage, tempLevelImage)


end subroutine RefineBackgroundImage1





!*****************************************************************
!
! RefineBackgroundImage
! Interpolates background across extended-source regions
!
!*****************************************************************
subroutine RefineBackgroundImage(image)
    real(single), pointer :: image(:, :)
  

    integer(int16) :: i, j, nx, ny, iu, ju, dx  
    real(single), pointer :: tempImage(:, :)
   
  
    write(errstr,'(a)') "Refining background  image"
    call message(AppMsg, Verbose)


    nx = ubound(image, 1)
    ny = ubound(image, 2)

    tempImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempimage")
    tempimage = image


    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    dx = 200 / binx
    do i = 0, nx, dx

        iu = min(i + dx, nx)
        do j = 0, ny, dx
            ju = min(j + dx, ny)


 !           write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Refining background  strip ", i, ":", iu, ",", j, ":", ju
 !           call message(AppMsg, Verbose)

            
            call RefineBackgroundStrip(i, iu, j, ju, image, tempImage)
            

        end do


    end do

!    call Kuwaharafilter(tempImage)
!    call Lpdfilter(tempImage)


!    call smoothimage( tempImage, int(nx + 1, int32), int(ny + 1, int32), &
!          int(20 / binx, int32), int(20 / binx, int32))

        
!call saveImage(image, nx + 1_int16, ny + 1_int16, "image.fits", "background image")
!call saveImage(tempimage, nx + 1_int16, ny + 1_int16, "tempimage.fits", "background image")


!call fatal("Line 1068 displaying image.fits and tempimage.fits")

    image = tempImage
    deallocate(tempimage)
    nullify(tempimage)


end subroutine RefineBackgroundImage
!*****************************************************************
!
!
!****************************************************************
subroutine RefineBackgroundStrip(il, iu, jl, ju, image, refinedImage)
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :), refinedImage(:, :)

    real(single) :: globalMode, sigma 
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int16) :: i, j
    integer(int16) :: dx
    integer(int32) :: n
    real(single), pointer, dimension(:) :: x=>null(), y=>null(), f=>null()

    dx = 200 / binx
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    il1 = max(0_int16, il - dx)
    iu1 = min(nx, iu + dx)
    jl1 = max(0_int16, jl - dx)
    ju1 = min(ny, ju + dx)



    n = (iu1 - il1 + 1) * (ju1 - jl1 + 1)

    x => allocateSingle(1, n, "x")
    y => allocateSingle(1, n, "y")
    f => allocateSingle(1, n, "f")
  

    globalMode = ComputeBackgroundMode3(image(il1 : iu1, jl1 : ju1), sigma)

!    write(*, *) "Line 980 global mode=", globalMode, " sigma=", sigma
    globalMode = globalMode + max(1.0, 0.25 * sigma)


    n = 0
    do j = jl, ju
        do i = il, iu
            if (image(i, j) .lt. globalMode .and. image(i, j) .ge. 0.0) then
                n = n + 1
                x(n) = real(i)
                y(n) = real(j)
                f(n) = image(i, j)
            end if
        end do
    end do

    do j = jl, ju
        do i = il, iu
            if (image(i, j) .ge. globalMode) then
                refinedimage(i, j) = NeighbourInterpolate(real(i), real(j), n, x, y, f)
            end if
        end do
    end do
     
    deallocate(x, y, f)
    nullify(x, y, f)

 
end subroutine RefineBackgroundStrip

!*****************************************************************
!
! RemoveSources
!
!*****************************************************************
subroutine RemoveSources(image, plevelimage, eLevelImage, source, nSources)
    real(single), pointer :: image(:, :)
    integer(int16), pointer, dimension(:, :) :: elevelImage, plevelImage

    type(sourceParameters), pointer :: source(:)

    integer(int16), intent(in) :: nSources
    integer(int16) :: i, j, k, nx, ny
    integer(int32) :: n 
    integer(int8) :: dx = 4, dx1
    real(single), pointer, dimension(:) :: x => null(), y => null(), f => null()
    integer(int16), pointer, dimension(:, :) :: etemplevelImage => null()
    integer(int16), pointer, dimension(:, :) :: ptemplevelImage => null()


    if (nSources .eq. 0) return
!    write(errstr,'(a)') "Removing sources"
!    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ptemplevelimage => allocate2dint16(0_int16, nx, 0_int16, ny, "ptemplevelimage")

    ptemplevelimage = plevelimage
    etemplevelimage => allocate2dint16(0_int16, nx, 0_int16, ny, "etemplevelimage")
    etemplevelimage = elevelimage
 
    dx = 4 / binx  ! was 8 
    call DilateImage(ptemplevelimage, dx)
    call DilateImage(etemplevelimage, dx)

    dx = 10 / binx  ! was 20 
    dx1 = 8 / binx

    n = (maxVal(source%imax - source%imin) + 2 * dx + 1) * &
        (maxVal(source%jmax - source%jmin) + 2 * dx + 1)

    x => allocateSingle(1, n, "x")
    y => allocateSingle(1, n, "y")
    f => allocateSingle(1, n, "f")
  
    ! Loop through all the sources and remove the sources from
    ! the image
    do i = 1, nSources 

!write(*, *) "Source ", i, " x=", source(i)%xpos, " y=", source(i)%ypos

!        if (source(i)%noSourcePixels .lt. 5) cycle
  
        n = 0


        do k = max(0_int16, source(i)%jmin - dx), min(ny, source(i)%jmax + dx)
            do j = max(0_int16, source(i)%imin - dx), min(nx, source(i)%imax + dx)
         
                if(ptemplevelImage(j, k) .ne. 0) cycle ! Exclude point-source pixels
                if(etemplevelImage(j, k) .ne. 0) cycle ! Exclude extended-source pixels
              
                n = n + 1
                x(n) = real(j)
                y(n) = real(k)
                f(n) = image(j, k)



            end do
        end do

        if (n .lt. 2) cycle

        do k = max(0_int16, source(i)%jmin - dx1), min(ny, source(i)%jmax + dx1)
            do j = max(0_int16, source(i)%imin - dx1), min(nx, source(i)%imax + dx1)
  
        
        
                   
                if(ptemplevelImage(j, k) .eq. 0) cycle
            
                image(j, k) = NeighbourInterpolate(real(j), real(k), n, x, y, f)

            end do
        end do
             
    end do

   


    call smoothimage(image, int(nx + 1, int32), int(ny + 1, int32), &
            int(6 / binx, int32), int(6 / binx, int32))



!    call saveImage(image, nx + 1_int16, ny + 1_int16, "backgroundimage.fits", "background image")

!call saveImage(ptemplevelimage, nx + 1_int16, ny + 1_int16, "ptemplevelimage.fits", "background image")
!call saveImage(etemplevelimage, nx + 1_int16, ny + 1_int16, "etemplevelimage.fits", "background image")

!call fatal("Line 1296 backgroundimage  displaying backgroundimage.fits and ptemplevelimage.fits and etemplevelimage.fits")


    deallocate(x, y, f, ptemplevelimage, etemplevelimage)
    nullify(x, y, f, ptemplevelimage, etemplevelimage)

 
end subroutine RemoveSources


!*****************************************************************
!
! RemovePointSources
! Purpose- Removes point-sources from an image that lie within
! an extended-source region
!
!*****************************************************************
subroutine RemovePointSources(image, plevelimage, eLevelImage, source, nSources)
    real(single), pointer :: image(:, :)
    integer(int16), pointer, dimension(:, :) :: elevelImage, plevelImage

    type(sourceParameters), pointer :: source(:)

    integer(int16), intent(in) :: nSources
    integer(int16) :: i, j, k, nx, ny
    integer(int32) :: n 
    integer(int8) :: dx = 4
    real(single), pointer, dimension(:) :: x => null(), y => null(), f => null()
  
    integer(int16), pointer, dimension(:, :) :: ptemplevelImage => null()


    if (nSources .eq. 0) return
    write(errstr,'(a)') "Removing point sources from extended-source regions"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    ptemplevelimage => allocate2dint16(0_int16, nx, 0_int16, ny, "ptemplevelimage")

    ptemplevelimage = plevelimage
    
 
    dx = 4 / binx ! was 4 binx
    call DilateImage(ptemplevelimage, dx)


    dx = 10 / binx 
 

    n = (maxVal(source%imax - source%imin) + 2 * dx + 1) * &
        (maxVal(source%jmax - source%jmin) + 2 * dx + 1)

    x => allocateSingle(1, n, "x")
    y => allocateSingle(1, n, "y")
    f => allocateSingle(1, n, "f")
  
    ! Loop through all the sources and remove the sources from the image

    do i = 1, nSources 


        ! Only remove point-sources from image that lie within an extended-source region
        if (.not. any(plevelImage .gt. 0_int16 .and. elevelImage .eq. source(i)%number) ) cycle


        n = 0


        do k = max(0_int16, source(i)%jmin - dx), min(ny, source(i)%jmax + dx)
            do j = max(0_int16, source(i)%imin - dx), min(nx, source(i)%imax + dx)
         
                if(ptemplevelImage(j, k) .ne. 0_int16) cycle ! Exclude point-source pixels
                
              
                n = n + 1
                x(n) = real(j)
                y(n) = real(k)
                f(n) = image(j, k)

            end do
        end do


        if (n .lt. 2) cycle

        do k = source(i)%jmin, source(i)%jmax 
            do j = source(i)%imin, source(i)%imax
  
                                
                if(plevelImage(j, k) .eq. 0_int16) cycle
                if(elevelImage(j, k) .eq. 0_int16) cycle ! Exclude point-source pixels
           
                image(j, k) = NeighbourInterpolate(real(j), real(k), n, x, y, f)

            end do
        end do
             
    end do

   
!call saveImage(image, nx + 1_int16, ny + 1_int16, "backgroundimage.fits", "background image")
!call saveImage(ptemplevelimage, nx + 1_int16, ny + 1_int16, "ptemplevelimage.fits", "background image")
!call saveImage(etemplevelimage, nx + 1_int16, ny + 1_int16, "etemplevelimage.fits", "background image")

!call fatal("Line 1296 backgroundimage  displaying backgroundimage.fits and ptemplevelimage.fits and etemplevelimage.fits")


    deallocate(x, y, f, ptemplevelimage)
    nullify(x, y, f, ptemplevelimage)



         
end subroutine RemovePointSources





!****************************************************************************
! Fill in the missing background pixels by interpolation
!***************************************************************************
subroutine FillInBackgroundPixels( image, sigmaImage, computed)
    real(single), pointer, dimension(:, :) :: image, sigmaImage 
    logical, pointer :: computed(:, :)
    type(PixelStruct), pointer :: pixel(:) => null()

    integer(int8) :: dl
    integer(int16) :: i, j, nx, ny


    dl = 3
    allocate( pixel( (2 * dl + 1) * (2 * dl + 1) ))

    nx = ubound(image, 1)
    ny = ubound(image, 2)



    do j = 0, ny        
        do i = 0, nx 
             
            if ( .not. computed(i, j) ) then
                image(i, j) = pixelAverage(i, j,   &
                              image,  sigmaImage,  &
                              pixel, dl, computed, &
                              sigmaImage(i, j) )                 

            end if 

        end do
    end do

    deallocate(pixel)
    nullify(pixel)
    

end subroutine FillInBackgroundPixels




end module backgroundimage_mod













