!*************************************************************
! Module sourcedetection_mod
! Created 15th April 2002- C.Brindle
! Contains a subroutine for detecting "point" sources and one 
! for detecting any other kind of source
!****************************************************************
module centroid_mod
    use cal
    use dal     
    use OmAllocation_mod     ! For allocating memory for arrays
    use OmSourceParam_mod    ! Defines structures for storing source parameters
    use omminim_mod
    

real(single), pointer, private :: imagePtr(:, :) => null()
real(single), private :: cRadius, cradius1
integer(int16), private :: cil, ciu, cjl, cju, il, iu, jl, ju
contains
!**************************************************************************************
!
!*************************************************************************************
subroutine Centroid(source, image, radius, binx)
    type(SourceParameters), intent(inout) :: source
    real(single), pointer :: image(:, :)
    real(single), intent(in) :: radius
    integer, intent(in) :: binx
    
    integer(int16) :: nx, ny
    real(double), dimension(2) :: x, step
    real(double) :: stopcr, simp, var(2), functionValue 
    integer(int8) :: dx
    integer(int16) :: i0, j0    
    integer :: nop, maxfn, nloop, iquad, ifault


    if (.not. associated(image)) return
   

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)

    cRadius = radius
    dx = cRadius + 2 
    cil = max(0_int16, i0 - dx)
    ciu = min(nx, i0 + dx)
    cjl = max(0_int16, j0 - dx)
    cju = min(ny, j0 + dx)

    cRadius1 = radius + real(8 / binx) 
    dx = cRadius + 2  

     
    il = max(0_int16, i0 - dx)
    iu = min(nx, i0 + dx)
    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)








   

    imagePtr => image

  
    step = 0.1_double
    nop = 2
    maxfn = 1000
    stopcr = 0.1_double
    nloop = 2 * nop
    iquad = 1
    x(1) = source%xpos - 1.0
    x(2) = source%ypos - 1.0
    simp = 0.0_double

    call omminim(x, step, nop, functionValue, maxfn, stopcr, nloop, iquad,  &
                 simp, var, MinimumFunction, ifault)


    x = x + 1.0_double

    write(*, *) "Function value=", functionValue, " ifault=", ifault
    write(*, *) "Initial x coord=", source%xpos, " centroided=", x(1)
    write(*, *) "Initial y coord=", source%ypos, " centroided=", x(2)

    source%xpos = x(1)
    source%ypos = x(2)

    nullify(imagePtr)


end subroutine Centroid

!**************************************************************************************
!
!*************************************************************************************
Function CentroidFunction(x) result(counts)
    real(double), intent(inout) :: x(:)   
    real(double) :: counts
    integer(int16) :: i, j, n
    real(double) :: dx, dy, radius


    counts = 0.0_double
    n = 0
    do j = cjl, cju
        dy = x(2) - dble(j)
        if (abs(dy) .gt. cRadius) cycle
        do i = cil, ciu
            dx = x(1) - dble(i)
            radius = sqrt(dx * dx + dy * dy)
            if (radius .le. cRadius) then
                counts = counts + dble(imagePtr(i, j)) * (cRadius - radius) / cRadius
                n = n + 1 
            end if

        end do
    end do

    if (n .gt. 0) then
        counts = -counts / dble(n)
    end if
end Function CentroidFunction


!**************************************************************************************
!
!*************************************************************************************
Function MinimumFunction(x) result(counts)
    real(double), intent(inout) :: x(:)   
    real(double) :: counts, counts1
    integer(int32) :: i, j, n, n1
    real(double) :: dx, dy, radius
     

    counts = 0.0_double

    counts1 = 0.0_double



    n = 0
    n1 = 0
    do j = il, iu
        dy = x(2) - dble(j)
        if (abs(dy) .gt. cRadius1) cycle
        do i = jl, ju
            dx = x(1) - dble(i)
            if (abs(dx) .gt. cRadius1) cycle
    
            radius = sqrt(dx * dx + dy * dy)
            if (radius .le. cRadius) then
                if (imagePtr(i, j) >= 0.0) then 
                    counts = counts + dble(imagePtr(i, j))
                    n = n + 1
                end if
            elseif(radius .le. cradius1) then
                if (imagePtr(i, j) >= 0.0) then 
                    counts1 = counts1 + dble(imagePtr(i, j))
                    n1 = n1 + 1
                end if
            


     
            end if

        end do
    end do

    if (n .gt. 0 .and. n1 > 0) then

        counts = -(counts1 / dble(n1) - counts / dble(n))

   

    end if







end Function MinimumFunction









end module centroid_mod


